# STDP-Tutorial: Spike Timing-Dependent Plasticity Learning Tutorial

| Task | Content | Status |
|----| ---- | ---- |
| Mathematical equations | Write the basic theory and the mathematical equations for the STDP learning | ✅ |
| Python | Write Python code; simulate and illustrate | ✅ |
| Verilog HDL | Write Verilog HDL of the neuron and testbench; simulate and illustrate | ⏳ |

This tutorial demonstrates how to implement Spike-Timing Dependent Plasticity (STDP) with Leaky Integrate-and-Fire (LIF) neurons to recognize a 3x3 binary pattern. The implementation includes Verilog HDL for hardware modeling and Python for simulation and visualization.

---

## 1. Introduction

### What is STDP?
Spike-Timing Dependent Plasticity (STDP) is a biological learning rule where the synaptic weights are updated based on the relative timing of spikes between pre-synaptic and post-synaptic neurons:
- **Pre-before-post**: Synapse strengthens (long-term potentiation, LTP).
- **Post-before-pre**: Synapse weakens (long-term depression, LTD).

### Objective
We will implement an STDP mechanism on an LIF neuron model to recognize a specific **3x3 binary pattern**. The goal is to adjust synaptic weights such that the neuron spikes predominantly when the target pattern is presented.

---

## 2. System Overview

### Components
1. **Input Layer**: 9 inputs (for the 3x3 grid).
2. **Synaptic Weights**: Adjustable weights connecting inputs to the LIF neuron.
3. **LIF Neuron**: Integrates inputs and generates spikes.
4. **STDP Mechanism**: Updates weights based on spike timing.

---


## 3. Python simulation

The following is the script for simulating STDP in Python. You can access it at [./sw/STDP.py](./sw/STDP.py).

```python
import numpy as np
import matplotlib.pyplot as plt

# Parameters
n_inputs = 9
v_threshold = 30
v_reset = 0
tau = 15
learning_rate = 0.01

# Initial conditions
weights = np.ones(n_inputs) * 0.5
membrane_potential = 0

# Input patterns
patterns = [
    np.array([1, 0, 1, 0, 1, 0, 1, 0, 1]),  # Target pattern
    np.array([0, 0, 0, 1, 1, 1, 0, 0, 0]),  # Noise
]

# Store initial weights for plotting
initial_weights = weights.copy()


# Storage for visualization
v_trace = []
weights_trace = []

# Simulation
for step in range(100):
    # Present pattern
    if step % 20 < 10:
        input_spikes = patterns[0]  # Target pattern
    else:
        input_spikes = patterns[1]  # Noise
    if step == 50:
        ts50_weights = weights.copy()
    # Integrate
    membrane_potential += np.dot(weights, input_spikes)
    v_trace.append(membrane_potential)
    weights_trace.append(weights.copy())

    # Spike condition
    if membrane_potential >= v_threshold:
        membrane_potential = v_reset

        # STDP update
        for i in range(n_inputs):
            if input_spikes[i] == 1:  # Pre-before-post
                weights[i] += learning_rate
            else:  # Post-before-pre
                weights[i] -= learning_rate

    # Decay potential
    membrane_potential -= membrane_potential / tau

initial_weights_3x3 = initial_weights.reshape((3, 3))
final_weights_3x3 = weights.reshape((3, 3))
ts50_weights_3x3 = ts50_weights.reshape((3,3))
# Plot results
fig1, axes1 = plt.subplots(2, 1, figsize=(10, 6))  


axes1[0].plot(v_trace, label="Membrane Potential")
axes1[0].axhline(y=v_threshold, color="r", linestyle="--", label="Threshold")
axes1[0].set_ylabel("Voltage")
axes1[0].set_xlabel("Timesteps")
axes1[0].legend()

axes1[1].plot(np.array(weights_trace), label=[f"Weight {i}" for i in range(n_inputs)])

axes1[1].set_ylabel("Synaptic strength")
axes1[1].set_xlabel("Timesteps")

axes1[0].legend()


fig2, axes2 = plt.subplots(2, 2, figsize=(10, 8))  

# # Pattern 0
pat0 = axes2[0, 0].imshow(patterns[0].reshape((3, 3)), cmap="viridis", interpolation="nearest")
fig2.colorbar(pat0, ax=axes2[0,0], label="Weight Value")
axes2[0, 0].set_title("Pattern 0 (2D Receiptive Field)")
axes2[0, 0].axis("off")

# # Pattern1
pat1 = axes2[0, 1].imshow(patterns[1].reshape((3, 3)), cmap="viridis", interpolation="nearest")
fig2.colorbar(pat1, ax=axes2[0,1], label="Weight Value")

axes2[0, 1].set_title("Pattern 1 (2D Receiptive Field)")
axes2[0, 1].axis("off")


# # Initial weights
bf0 = axes2[1, 0].imshow(initial_weights_3x3, cmap="viridis", interpolation="nearest")
fig2.colorbar(bf0,  ax=axes2[1,0], label="Weight Value")
axes2[1, 0].set_title("Initial Weights  (2D Receiptive Field)")
axes2[1, 0].axis("off")

# # Final weights
af0 = axes2[1, 1].imshow(final_weights_3x3, cmap="viridis", interpolation="nearest")
fig2.colorbar(af0,  ax=axes2[1,1], label="Weight Value")
axes2[1, 1].set_title("Final Weights  (2D Receiptive Field)")
axes2[1, 1].axis("off")

plt.tight_layout()

plt.show()
```

### Illustration

#### These are the plots after 100 timesteps

The plot of the membrane potentail and the weight are shown below:

![Waveform of membrane potential and 9 weights.](./imgs/sw_1.png)

The 2D receiptive field can be observed as follow:
![The 2D receiptive field of: pattern 0, pattern1, initialized weight and final weight.](./imgs/sw_2.png)

Observation:

- There are two patterns (#0 and #1) are fed into the neuron solely.
- The neuron catches the pattern 0 and change its weight using STDP learning.
- The final weight is in a similar state as pattern 0; however, the level of intensity are different. On the weights, their values range between 0.480 and 0.520.
- The low intensity is due to the simulation is only 100 time steps. The neuron fired only twice which make the weight changes insignificant.

#### These are the plots after 1000 timesteps


The plot of the membrane potentail and the weight are shown below:

![Waveform of membrane potential and 9 weights.](./imgs/sw_3.png)

The 2D receiptive field can be observed as follow:
![The 2D receiptive field of: pattern 0, pattern1, initialized weight and final weight.](./imgs/sw_4.png)

Observation:

- By increasing the number of time-steps, the neuron fires a lot more.
- By firing more, the neuron keep updating its weight and increase the intensity of it weight.
  
## 4. Verilog HDL Implementation

### LIF Neuron with Synapses and STDP
```verilog
module LIF_Neuron_STDP (
    input clk,
    input reset,
    input [8:0] input_spikes, // 3x3 grid (flattened)
    output reg spike_out
);

    // Parameters
    parameter signed [15:0] v_threshold = 16'sd30;
    parameter signed [15:0] v_reset = 16'sd0;
    parameter signed [15:0] tau = 16'sd15; // Membrane decay rate

    // Internal variables
    reg signed [15:0] membrane_potential;
    reg signed [7:0] weights [0:8]; // Synaptic weights (9 inputs)

    // Initialize weights
    integer i;
    initial begin
        for (i = 0; i < 9; i = i + 1) begin
            weights[i] = 8'sd10; // Initial weight for each synapse
        end
    end

    // STDP update variables
    reg [8:0] pre_spike_times; // Time since last pre-spike for each input
    reg post_spike_time;       // Time since last post-spike

    // Spike Timing Update
    always @(posedge clk or posedge reset) begin
        if (reset) begin
            membrane_potential <= 0;
            spike_out <= 0;
            pre_spike_times <= 0;
            post_spike_time <= 0;
        end else begin
            // Decay membrane potential
            membrane_potential <= membrane_potential - membrane_potential / tau;

            // Integrate inputs
            for (i = 0; i < 9; i = i + 1) begin
                if (input_spikes[i]) begin
                    membrane_potential <= membrane_potential + weights[i];
                    pre_spike_times[i] <= 0; // Reset pre-spike timer
                end else begin
                    pre_spike_times[i] <= pre_spike_times[i] + 1;
                end
            end

            // Spike condition
            if (membrane_potential >= v_threshold) begin
                spike_out <= 1;
                membrane_potential <= v_reset; // Reset membrane potential
                post_spike_time <= 0;         // Reset post-spike timer

                // STDP weight update
                for (i = 0; i < 9; i = i + 1) begin
                    if (pre_spike_times[i] < 16) begin
                        // LTP: Pre-before-post
                        weights[i] <= weights[i] + 1;
                    end else if (post_spike_time < 16) begin
                        // LTD: Post-before-pre
                        weights[i] <= weights[i] - 1;
                    end
                end
            end else begin
                spike_out <= 0;
                post_spike_time <= post_spike_time + 1;
            end
        end
    end
endmodule
```

### Testbench

```verilog
`timescale 1ns/1ps

module LIF_Neuron_STDP_tb;

    reg clk;
    reg reset;
    reg [8:0] input_spikes;
    wire spike_out;

    // Instantiate the DUT
    LIF_Neuron_STDP dut (
        .clk(clk),
        .reset(reset),
        .input_spikes(input_spikes),
        .spike_out(spike_out)
    );

    // Clock generation
    always #5 clk = ~clk;

    // Test sequence
    initial begin
        $dumpfile("waveform.vcd");
        $dumpvars(0, LIF_Neuron_STDP_tb);

        clk = 0;
        reset = 1;
        input_spikes = 9'b0;

        #10 reset = 0;

        // Present target pattern
        input_spikes = 9'b101010101; // Example 3x3 pattern
        #100;

        // Random noise
        input_spikes = 9'b000111000;
        #100;

        // Repeat target pattern
        input_spikes = 9'b101010101;
        #100;

        $stop;
    end
endmodule
```

