module LIF_Neuron_STDP (
    input clk,
    input reset,
    input [8:0] input_spikes, // 3x3 grid (flattened)
    output reg spike_out
);

    // Parameters
    parameter signed [15:0] v_threshold = 16'sd30;
    parameter signed [15:0] v_reset = 16'sd0;
    parameter signed [15:0] tau = 16'sd15; // Membrane decay rate

    // Internal variables
    reg signed [15:0] membrane_potential;
    reg signed [7:0] weights [0:8]; // Synaptic weights (9 inputs)

    // Initialize weights
    integer i;
    initial begin
        for (i = 0; i < 9; i = i + 1) begin
            weights[i] = 8'sd10; // Initial weight for each synapse
        end
    end

    // STDP update variables
    reg [8:0] pre_spike_times; // Time since last pre-spike for each input
    reg [8:0] post_spike_time;       // Time since last post-spike

    // Spike Timing Update
    always @(posedge clk or posedge reset) begin
        if (reset) begin
            membrane_potential <= 0;
            spike_out <= 0;
            pre_spike_times <= 0;
            post_spike_time <= 0;
        end else begin
            // Decay membrane potential
            membrane_potential <= membrane_potential - membrane_potential / tau;

            // Integrate inputs
            for (i = 0; i < 9; i = i + 1) begin
                if (input_spikes[i]) begin
                    membrane_potential <= membrane_potential + weights[i];
                    pre_spike_times[i] <= 0; // Reset pre-spike timer
                end else begin
                    pre_spike_times[i] <= pre_spike_times[i] + 1;
                end
            end

            // Spike condition
            if (membrane_potential >= v_threshold) begin
                spike_out <= 1;
                membrane_potential <= v_reset; // Reset membrane potential
                post_spike_time <= 0;         // Reset post-spike timer

                // STDP weight update
                for (i = 0; i < 9; i = i + 1) begin
                    if (pre_spike_times[i] == 1'b1) begin
                        // LTP: Pre-before-post
                        if (weights[i] != 8'b11111111) begin
                            weights[i] <= weights[i] + 1;
                        end
                    end else if (post_spike_time == 1'b1) begin
                        // LTD: Post-before-pre
                        if (weights[i] != 8'b00000000) begin
                            weights[i] <= weights[i] - 1;
                        end
                    end
                end
            end else begin
                spike_out <= 0;
                post_spike_time <= post_spike_time + 1;
            end
        end
    end
endmodule
