import numpy as np
import matplotlib.pyplot as plt

class LIFNeuron:
    def __init__(self, tau_m=20.0, R=1.0, V_th=1.0, V_reset=0.0, dt=1.0):
        self.tau_m = tau_m  # Membrane time constant
        self.R = R  # Membrane resistance
        self.V_th = V_th  # Spike threshold
        self.V_reset = V_reset  # Reset potential
        self.dt = dt  # Time step
        self.V_m = 0.0  # Membrane potential
        self.membrane_potential = []  # To store membrane potential over time
        self.spikes = []  # To store spike times

    def update(self, I):
        """
        Update the membrane potential based on input current I.
        Returns True if the neuron spikes, False otherwise.
        """
        dV = (self.dt / self.tau_m) * (-(self.V_m - self.V_reset) + self.R * I)
        self.V_m += dV

        # Store membrane potential
        self.membrane_potential.append(self.V_m)

        if self.V_m >= self.V_th:
            self.V_m = self.V_reset
            self.spikes.append(len(self.membrane_potential))  # Store spike time
            return True
        else:
            return False

    def inhibit(self, inhibition):
        """
        Apply lateral inhibition to reduce the membrane potential.
        """
        self.V_m -= inhibition

    def reset(self):
        """Reset the neuron for the next trial."""
        self.V_m = self.V_reset
        self.membrane_potential = []
        self.spikes = []


class STDP:
    def __init__(self, num_neurons, tau_pre=20.0, tau_post=20.0, A_plus=0.005, A_minus=0.005):
        self.num_neurons = num_neurons
        self.tau_pre = tau_pre
        self.tau_post = tau_post
        self.A_plus = A_plus
        self.A_minus = A_minus
        
        # Initialize synaptic weights with random values, including inhibitory connections
        self.weights = np.random.uniform(-0.5, 0.5, (num_neurons, num_neurons))
        
        # Initialize pre- and post-synaptic traces
        self.pre_trace = np.zeros(num_neurons)
        self.post_trace = np.zeros(num_neurons)

    def update_weights(self, pre_spikes, post_spikes, dt):
        # Update pre- and post-synaptic traces
        self.pre_trace = self.pre_trace * np.exp(-dt / self.tau_pre) + pre_spikes
        self.post_trace = self.post_trace * np.exp(-dt / self.tau_post) + post_spikes

        for i in range(self.num_neurons):
            for j in range(self.num_neurons):
                if pre_spikes[i] and post_spikes[j]:  # Both pre- and post-synaptic neurons spiked
                    delta_t = post_spikes[j] - pre_spikes[i]
                    if delta_t > 0:
                        self.weights[i, j] += self.A_plus * np.exp(-delta_t / self.tau_pre)
                    else:
                        self.weights[i, j] -= self.A_minus * np.exp(delta_t / self.tau_post)

    def get_weights(self):
        return self.weights


class PatternRecognizer:
    def __init__(self, pattern_size, inhibition_strength=0.5):
        self.pattern_size = pattern_size
        self.num_neurons = pattern_size * pattern_size
        self.stdp = STDP(self.num_neurons)
        self.neurons = [LIFNeuron() for _ in range(self.num_neurons)]
        self.inhibition_strength = inhibition_strength  # Strength of lateral inhibition

    def train(self, pattern, epochs=100, dt=1.0):
        pattern_flat = pattern.flatten()
        for _ in range(epochs):
            pre_spikes = pattern_flat
            post_spikes = np.zeros(self.num_neurons)

            # Reset neurons for each epoch
            for neuron in self.neurons:
                neuron.reset()

            # Update each neuron's membrane potential and check for spikes
            for t in range(int(100 / dt)):  # Simulate for 100 time steps
                for i in range(self.num_neurons):
                    I = np.dot(self.stdp.get_weights()[i], pre_spikes)
                    if self.neurons[i].update(I):
                        post_spikes[i] = t
                        # Apply lateral inhibition to other neurons
                        for j in range(self.num_neurons):
                            if i != j:
                                self.neurons[j].inhibit(self.inhibition_strength)

            self.stdp.update_weights(pre_spikes, post_spikes, dt)

    def recognize(self, pattern, dt=1.0):
        pre_spikes = pattern.flatten()
        post_spikes = np.zeros(self.num_neurons)

        # Reset neurons for recognition
        for neuron in self.neurons:
            neuron.reset()

        # Update each neuron's membrane potential and check for spikes
        for t in range(int(100 / dt)):  # Simulate for 100 time steps
            for i in range(self.num_neurons):
                I = np.dot(self.stdp.get_weights()[i], pre_spikes)
                if self.neurons[i].update(I):
                    post_spikes[i] = t
                    # Apply lateral inhibition to other neurons
                    for j in range(self.num_neurons):
                        if i != j:
                            self.neurons[j].inhibit(self.inhibition_strength)

        return post_spikes

    def get_weights(self):
        return self.stdp.get_weights()

    def get_neurons(self):
        return self.neurons


def plot_pattern(pattern, title="Pattern"):
    plt.imshow(pattern, cmap='gray', interpolation='nearest')
    plt.title(title)
    plt.colorbar()
    plt.show()

def plot_weights(weights, title="Weights"):
    plt.imshow(weights, cmap='viridis', interpolation='nearest')
    plt.title(title)
    plt.colorbar()
    plt.show()

def plot_individual_weights(weights, pattern_size):
    num_neurons = pattern_size * pattern_size
    fig, axs = plt.subplots(pattern_size, pattern_size, figsize=(10, 10))
    fig.suptitle("Individual Neuron Weights")
    for i in range(num_neurons):
        ax = axs[i // pattern_size, i % pattern_size]
        weight_matrix = weights[i].reshape((pattern_size, pattern_size))
        ax.imshow(weight_matrix, cmap='viridis', interpolation='nearest')
        ax.set_title(f"Neuron {i+1}")
        ax.axis('off')
    plt.tight_layout()
    plt.show()

def plot_membrane_potential_and_spikes(neurons, pattern_size):
    num_neurons = pattern_size * pattern_size
    fig, axs = plt.subplots(pattern_size, pattern_size, figsize=(10, 10))
    fig.suptitle("Membrane Potential and Spiking Activity")
    for i in range(num_neurons):
        ax = axs[i // pattern_size, i % pattern_size]
        ax.plot(neurons[i].membrane_potential, label='Membrane Potential')
        spike_times = neurons[i].spikes
        ax.plot(spike_times, [neurons[i].V_th] * len(spike_times), 'ro', label='Spikes')
        ax.set_title(f"Neuron {i+1}")
        ax.legend()
    plt.tight_layout()
    plt.show()

# Example 3x3 pattern
pattern = np.array([
    [1, 0, 1],
    [0, 1, 0],
    [1, 0, 1]
])

recognizer = PatternRecognizer(pattern_size=3)
recognizer.train(pattern, epochs=100)

# Plot the input pattern
plot_pattern(pattern, title="Input Pattern")

# Get and plot the learned weights
learned_weights = recognizer.get_weights()
plot_weights(learned_weights, title="Learned Weights")

# Plot individual neuron weights
plot_individual_weights(learned_weights, pattern_size=3)

# Plot membrane potential and spiking activities
neurons = recognizer.get_neurons()
plot_membrane_potential_and_spikes(neurons, pattern_size=3)