import numpy as np
import matplotlib.pyplot as plt

# Parameters
n_inputs = 9
v_threshold = 30
v_reset = 0
tau = 15
learning_rate = 0.01
weight_norm = True
# Initial conditions
weights = np.ones(n_inputs) * 0.5
membrane_potential = 0
fixed_weight_sum = np.sum(weights)

# Input patterns
patterns = [
    np.array([1, 0, 1, 0, 1, 0, 1, 0, 1]),  # Target pattern
    np.array([0, 0, 0, 1, 1, 1, 0, 0, 0]),  # Noise
]

# Store initial weights for plotting
initial_weights = weights.copy()


# Storage for visualization
v_trace = []
weights_trace = []

# Simulation
for step in range(1000):
    # Present pattern
    if step % 20 < 10:
        input_spikes = patterns[0]  # Target pattern
    else:
        input_spikes = patterns[1]  # Noise

    # Integrate
    membrane_potential += np.dot(weights, input_spikes)
    v_trace.append(membrane_potential)
    weights_trace.append(weights.copy())

    # Spike condition
    if membrane_potential >= v_threshold:
        membrane_potential = v_reset

        # STDP update
        for i in range(n_inputs):
            if input_spikes[i] == 1:  # Pre-before-post
                weights[i] += learning_rate
            else:  # Post-before-pre
                weights[i] -= learning_rate
    if weight_norm: # Weight normalization
        new_weight_sum = np.sum(weights)
        weight_scale_factor = new_weight_sum/fixed_weight_sum
        for i in range (n_inputs):
            weights[i] = weights[i]/weight_scale_factor

    # Decay potential
    membrane_potential -= membrane_potential / tau

initial_weights_3x3 = initial_weights.reshape((3, 3))
final_weights_3x3 = weights.reshape((3, 3))
print(np.max(final_weights_3x3))
print(np.min(final_weights_3x3))
print(fixed_weight_sum)
print(np.sum(weights))
# Plot results
fig1, axes1 = plt.subplots(2, 1, figsize=(10, 6))  


axes1[0].plot(v_trace, label="Membrane Potential")
axes1[0].axhline(y=v_threshold, color="r", linestyle="--", label="Threshold")
axes1[0].set_ylabel("Voltage")
axes1[0].set_xlabel("Timesteps")
axes1[0].legend()

axes1[1].plot(np.array(weights_trace), label=[f"Weight {i}" for i in range(n_inputs)])

axes1[1].set_ylabel("Synaptic strength")
axes1[1].set_xlabel("Timesteps")

axes1[0].legend()


fig2, axes2 = plt.subplots(2, 2, figsize=(10, 8))  

# # Pattern 0
pat0 = axes2[0, 0].imshow(patterns[0].reshape((3, 3)), cmap="viridis", interpolation="nearest")
fig2.colorbar(pat0, ax=axes2[0,0], label="Weight Value")
axes2[0, 0].set_title("Pattern 0 (2D Receiptive Field)")
axes2[0, 0].axis("off")

# # Pattern1
pat1 = axes2[0, 1].imshow(patterns[1].reshape((3, 3)), cmap="viridis", interpolation="nearest")
fig2.colorbar(pat1, ax=axes2[0,1], label="Weight Value")

axes2[0, 1].set_title("Pattern 1 (2D Receiptive Field)")
axes2[0, 1].axis("off")


# # Initial weights
bf0 = axes2[1, 0].imshow(initial_weights_3x3, cmap="viridis", interpolation="nearest")
fig2.colorbar(bf0,  ax=axes2[1,0], label="Weight Value")
axes2[1, 0].set_title("Initial Weights  (2D Receiptive Field)")
axes2[1, 0].axis("off")

# # Final weights
af0 = axes2[1, 1].imshow(final_weights_3x3, cmap="viridis", interpolation="nearest")
fig2.colorbar(af0,  ax=axes2[1,1], label="Weight Value")
axes2[1, 1].set_title("Final Weights  (2D Receiptive Field)")
axes2[1, 1].axis("off")

plt.tight_layout()

plt.show()
